import operator
import re
from pathlib import Path
from typing import ClassVar, Iterable

from pydantic import Field
from pydantic.config import ConfigDict
from pydantic.dataclasses import dataclass

from organize.filter import FilterConfig
from organize.output import Output
from organize.resource import Resource
from organize.validators import FlatList

OPERATORS = {
    "<": operator.lt,
    "<=": operator.le,
    "==": operator.eq,
    "=": operator.eq,
    "": operator.eq,
    ">=": operator.ge,
    ">": operator.gt,
}

SIZE_REGEX = re.compile(
    r"^(?P<op>[<>=]*)(?P<num>(\d*\.)?\d+)(?P<unit>[kmgtpezy]?i?)b?$"
)


def read_file_size(path: Path) -> int:
    return path.stat().st_size


def read_dir_size(path: Path) -> int:
    return sum(f.stat().st_size for f in path.glob("**/*") if f.is_file())


def read_resource_size(res: Resource) -> int:
    assert res.path is not None
    if res.is_file():
        return read_file_size(res.path)
    if res.is_dir():
        return read_dir_size(res.path)
    raise ValueError("Unknown file type")


def create_constraints(inp: str):
    """
    Given an input string it returns a list of tuples (comparison operator,
    number of bytes).

    Accepted formats are: "30k", ">= 5 TiB, <10tb", "< 60 tb", ...
    Calculation is in bytes, even if the "b" is lowercase. If an "i" is present
    we calculate base 1024.
    """
    parts = str(inp).replace(" ", "").lower().split(",")
    for part in parts:
        try:
            reg_match = SIZE_REGEX.match(part)
            if reg_match:
                match = reg_match.groupdict()
                op = OPERATORS[match["op"]]
                num = float(match["num"]) if "." in match["num"] else int(match["num"])
                unit = match["unit"]
                base = 1024 if unit.endswith("i") else 1000
                exp = "kmgtpezy".index(unit[0]) + 1 if unit else 0
                numbytes = num * base**exp
                yield (op, numbytes)
        except (AttributeError, KeyError, IndexError, ValueError, TypeError) as e:
            raise ValueError("Invalid size format: %s" % part) from e


def satisfies_constraints(size, constraints) -> bool:
    return all(op(size, p_size) for op, p_size in constraints)


def number_with_unit(size: int, suffixes: Iterable[str], base: int) -> str:
    size = int(size)
    if size == 1:
        return "1 byte"
    elif size < base:
        return "{:,} bytes".format(size)

    for i, suffix in enumerate(suffixes, 2):
        unit = base**i
        if size < unit:
            break
    return "{:,.1f} {}".format((base * size / unit), suffix)


def traditional(size: int) -> str:
    """Convert a filesize in to a string (powers of 1024, JDEC prefixes)."""
    return number_with_unit(
        size, ("KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB"), 1024
    )


def binary(size: int) -> str:
    """Convert a filesize in to a string (powers of 1024, IEC prefixes)."""
    return number_with_unit(
        size, ("KiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB", "YiB"), 1024
    )


def decimal(size: int) -> str:
    """Convert a filesize in to a string (powers of 1000, SI prefixes)."""
    return number_with_unit(
        size, ("kB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB"), 1000
    )


@dataclass(config=ConfigDict(coerce_numbers_to_str=True, extra="forbid"))
class Size:
    """Matches files and folders by size

    Attributes:
        *conditions (list(str) or str):
            The size constraints.

    Accepts file size conditions, e.g: `">= 500 MB"`, `"< 20k"`, `">0"`,
    `"= 10 KiB"`.

    It is possible to define both lower and upper conditions like this:
    `">20k, < 1 TB"`, `">= 20 Mb, <25 Mb"`. The filter will match if all given
    conditions are satisfied.

    - Accepts all units from KB to YB.
    - If no unit is given, kilobytes are assumend.
    - If binary prefix is given (KiB, GiB) the size is calculated using base 1024.

    **Returns:**

    - `{size.bytes}`: (int) Size in bytes
    - `{size.traditional}`: (str) Size with unit (powers of 1024, JDEC prefixes)
    - `{size.binary}`: (str) Size with unit (powers of 1024, IEC prefixes)
    - `{size.decimal}`: (str) Size with unit (powers of 1000, SI prefixes)
    """

    conditions: FlatList[str] = Field(default_factory=list)

    filter_config: ClassVar[FilterConfig] = FilterConfig(
        name="size", files=True, dirs=True
    )

    def __post_init__(self):
        self._constraints = set()
        for x in self.conditions:
            for constraint in create_constraints(x):
                self._constraints.add(constraint)

    def matches(self, filesize: int) -> bool:
        if not self._constraints:
            return True
        return all(op(filesize, c_size) for op, c_size in self._constraints)

    def pipeline(self, res: Resource, output: Output) -> bool:
        assert res.path is not None
        bytes = read_resource_size(res=res)
        res.vars[self.filter_config.name] = {
            "bytes": bytes,
            "traditional": traditional(bytes),
            "binary": binary(bytes),
            "decimal": decimal(bytes),
        }
        return self.matches(bytes)



import shutil
from typing import ClassVar

from pydantic.config import ConfigDict
from pydantic.dataclasses import dataclass

from organize.action import ActionConfig
from organize.output import Output
from organize.resource import Resource
from organize.template import Template, render

from .common.conflict import ConflictMode, resolve_conflict
from .common.target_path import prepare_target_path


@dataclass(config=ConfigDict(coerce_numbers_to_str=True, extra="forbid"))
class Move:

    """Move a file to a new location.

    The file can also be renamed.
    If the specified path does not exist it will be created.

    If you only want to rename the file and keep the folder, it is
    easier to use the `rename` action.

    Attributes:
        dest (str):
            The destination where the file / dir should be moved to.
            If `dest` ends with a slash, it is assumed to be a target directory
            and the file / dir will be moved into `dest` and keep its name.

        on_conflict (str):
            What should happen in case **dest** already exists.
            One of `skip`, `overwrite`, `trash`, `rename_new` and `rename_existing`.
            Defaults to `rename_new`.

        rename_template (str):
            A template for renaming the file / dir in case of a conflict.
            Defaults to `{name} {counter}{extension}`.

        autodetect_folder (bool):
            In case you forget the ending slash "/" to indicate moving into a folder
            this settings will handle targets without a file extension as folders.
            If you really mean to move to a file without file extension, set this to
            false.
            Default: True

    The next action will work with the moved file / dir.
    """

    dest: str
    on_conflict: ConflictMode = "rename_new"
    rename_template: str = "{name} {counter}{extension}"
    autodetect_folder: bool = True

    action_config: ClassVar[ActionConfig] = ActionConfig(
        name="move",
        standalone=False,
        files=True,
        dirs=True,
    )

    def __post_init__(self):
        self._dest = Template.from_string(self.dest)
        self._rename_template = Template.from_string(self.rename_template)

    def pipeline(self, res: Resource, output: Output, simulate: bool):
        assert res.path is not None, "Does not support standalone mode"
        rendered = render(self._dest, res.dict())

        # fully resolve the destination for folder targets and prepare the folder
        # structure
        dst = prepare_target_path(
            src_name=res.path.name,
            dst=rendered,
            autodetect_folder=self.autodetect_folder,
            simulate=simulate,
        )

        # Resolve conflicts before moving the file to the destination
        skip_action, dst = resolve_conflict(
            dst=dst,
            res=res,
            conflict_mode=self.on_conflict,
            rename_template=self._rename_template,
            simulate=simulate,
            output=output,
        )

        if skip_action:
            return

        output.msg(res=res, msg=f"Move to {dst}", sender=self)
        res.walker_skip_pathes.add(dst)
        if not simulate:
            shutil.move(src=res.path, dst=dst)

        # continue with the new path
        res.path = dst



import mimetypes
from typing import ClassVar

from pydantic import Field
from pydantic.config import ConfigDict
from pydantic.dataclasses import dataclass

from organize.filter import FilterConfig
from organize.output import Output
from organize.resource import Resource
from organize.validators import FlatList


def guess_mimetype(path):
    type_, _ = mimetypes.guess_type(path, strict=False)
    return type_


@dataclass(config=ConfigDict(coerce_numbers_to_str=True, extra="forbid"))
class MimeType:

    """Filter by MIME type associated with the file extension.

    Supports a single string or list of MIME type strings as argument.
    The types don't need to be fully specified, for example "audio" matches everything
    from "audio/midi" to "audio/quicktime".

    You can see a list of known MIME types on your system by running this oneliner:

    ```sh
    python3 -m organize.filters.mimetype
    ```

    Attributes:
        *mimetypes (list(str) or str): The MIME types to filter for.

    **Returns:**

    - `{mimetype}`: The MIME type of the file.
    """

    mimetypes: FlatList[str] = Field(default_factory=list)

    filter_config: ClassVar[FilterConfig] = FilterConfig(
        name="mimetype",
        files=True,
        dirs=False,
    )

    def matches(self, mimetype) -> bool:
        if mimetype is None:
            return False
        if not self.mimetypes:
            return True
        return any(mimetype.startswith(x) for x in self.mimetypes)

    def pipeline(self, res: Resource, output: Output) -> bool:
        mimetype = guess_mimetype(res.path)
        res.vars[self.filter_config.name] = mimetype
        return self.matches(mimetype)


if __name__ == "__main__":
    all_types = set(mimetypes.common_types.values()) | set(mimetypes.types_map.values())
    for t in sorted(all_types):
        print(t)



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Literal, Optional, Union

from pydantic import BaseModel

from ._sender import sender_name
from .output import Level

if TYPE_CHECKING:
    from organize.resource import Resource

    from ._sender import SenderType


class Start(BaseModel):
    type: Literal["START"] = "START"
    simulate: bool
    config_path: Optional[Path]
    working_dir: Path


class Msg(BaseModel):
    type: Literal["MSG"] = "MSG"
    level: Level = "info"
    path: Optional[Path]
    basedir: Optional[Path]
    sender: str
    msg: str
    rule_nr: int
    rule_name: str


class Confirm(BaseModel):
    type: Literal["CONFIRM"] = "CONFIRM"
    path: Optional[Path]
    basedir: Optional[Path]
    sender: str
    msg: str
    default: bool
    rule_nr: int
    rule_name: str


class Report(BaseModel):
    type: Literal["REPORT"] = "REPORT"
    success_count: int
    error_count: int


EventType = Union[Start, Msg, Confirm, Report]


class JSONL:
    def __init__(self, auto_confirm: bool = False) -> None:
        self.auto_confirm = auto_confirm

    def start(
        self,
        simulate: bool,
        config_path: Optional[Path],
        working_dir: Path,
    ) -> None:
        self.emit_event(
            Start(
                simulate=simulate,
                config_path=config_path.resolve() if config_path else None,
                working_dir=working_dir.resolve(),
            )
        )

    def msg(
        self,
        res: Resource,
        msg: str,
        sender: SenderType,
        level: Level = "info",
    ) -> None:
        self.emit_event(
            Msg(
                level=level,
                path=res.path,
                basedir=res.basedir,
                sender=sender_name(sender),
                msg=msg,
                rule_nr=res.rule_nr,
                rule_name=res.rule.name if res.rule and res.rule.name else "",
            )
        )

    def confirm(
        self,
        res: Resource,
        msg: str,
        default: bool,
        sender: SenderType,
    ) -> bool:
        if self.auto_confirm:
            return True
        self.emit_event(
            Confirm(
                path=res.path,
                basedir=res.basedir,
                sender=sender_name(sender),
                msg=msg,
                default=default,
                rule_nr=res.rule_nr,
                rule_name=res.rule.name if res.rule and res.rule.name else "",
            )
        )
        answer = input().lower()
        if answer == "":
            return default
        elif answer in ("j", "y", "ja", "yes", "1"):
            return True
        return False

    def end(self, success_count: int, error_count: int) -> None:
        report = Report(
            success_count=success_count,
            error_count=error_count,
        )
        self.emit_event(report)

    def emit_event(self, event: EventType) -> None:
        print(event.model_dump_json())



from pathlib import Path
from typing import Iterable, Optional

from pydantic import ValidationError


class ConfigError(ValueError):
    def __init__(self, e: ValidationError, config_path: Optional[Path] = None):
        self.e = e
        self.config_path = config_path.resolve() if config_path else None

    def __str__(self):
        count = self.e.error_count()
        title = self.config_path or self.e.title

        lines = [
            f'{count} validation error{"s" if count != 1 else ""} in "{title}"',
            "",
        ]
        for err in self.e.errors(include_url=False):
            loc = ".".join(str(x) for x in err["loc"])
            msg = err["msg"]
            inp_name = err["input"]
            inp_type = type(err["input"]).__name__
            inp = f'[input_value="{inp_name}", input_type={inp_type}]'
            lines.append(loc)
            lines.append(f"  {msg} {inp}")
        return "\n".join(lines)

    def json(self):
        return self.e.json()


class ConfigNotFound(FileNotFoundError):
    def __init__(
        self,
        config: str,
        search_pathes: Iterable[Path] = tuple(),
        init_path: Optional[Path] = None,
    ):
        self.config = config
        self.search_pathes = search_pathes
        self.init_path = init_path

    def __str__(self):
        msg = f'Cannot find config "{self.config}".'
        if self.search_pathes:
            path_listing = "\n".join(f' - "{path}"' for path in self.search_pathes)
            return f"{msg}\nSearch locations:\n{path_listing}"
        return msg



from __future__ import annotations

from typing import Dict, Type

from . import actions, filters
from .action import Action
from .filter import Filter

FILTERS: Dict[str, Type[Filter]] = dict()
ACTIONS: Dict[str, Type[Action]] = dict()


def register_filter(filter: Type[Filter], force: bool = False):
    name = filter.filter_config.name
    if not force and name in FILTERS:
        raise ValueError(f'"{name}" is already registered for filter {FILTERS[name]}')
    FILTERS[name.lower()] = filter


def filter_by_name(name: str) -> Type[Filter]:
    try:
        return FILTERS[name.lower()]
    except KeyError as e:
        raise ValueError(f'Unknown filter: "{name}"') from e


def register_action(action: Type[Action], force: bool = False):
    name = action.action_config.name
    if not force and name in ACTIONS:
        raise ValueError(f'"{name}" is already registered for action {ACTIONS[name]}')
    ACTIONS[name.lower()] = action


def action_by_name(name: str) -> Type[Action]:
    try:
        return ACTIONS[name.lower()]
    except KeyError as e:
        raise ValueError(f'Unknown action: "{name}"') from e


# Register filters and actions
for x in filters.ALL:
    register_filter(x)
for x in actions.ALL:
    register_action(x)



import re
from typing import ClassVar

from pydantic.config import ConfigDict
from pydantic.dataclasses import dataclass

from organize.filter import FilterConfig
from organize.output import Output
from organize.resource import Resource
from organize.utils import normalize_unicode


@dataclass(config=ConfigDict(coerce_numbers_to_str=True, extra="forbid"))
class Regex:

    """Matches filenames with the given regular expression

    Attributes:
        expr (str): The regular expression to be matched.

    **Returns:**

    Any named groups in your regular expression will be returned like this:

    - `{regex.groupname}`: The text matched with the named
      group `(?P<groupname>.*)`

    """

    expr: str

    filter_config: ClassVar[FilterConfig] = FilterConfig(
        name="regex",
        files=True,
        dirs=True,
    )

    def __post_init__(self):
        self._expr = re.compile(self.expr, flags=re.UNICODE)

    def matches(self, path: str):
        return self._expr.search(path)

    def pipeline(self, res: Resource, output: Output) -> bool:
        assert res.path is not None, "Does not support standalone mode"
        match = self.matches(normalize_unicode(res.path.name))
        if match:
            res.deep_merge(key=self.filter_config.name, data=match.groupdict())
            return True
        return False



import subprocess
import sys
from datetime import datetime
from pathlib import Path
from typing import ClassVar

from organize.filter import FilterConfig

from .common.timefilter import TimeFilter


def read_date_lastused(path: Path) -> datetime:
    cmd = ["mdls", "-name", "kMDItemLastUsedDate", "-raw", str(path)]
    out = subprocess.check_output(cmd, encoding="utf-8").strip()
    if out == "(null)":
        raise ValueError("date_lastused not available")
    return datetime.strptime(out, "%Y-%m-%d %H:%M:%S %z")


class DateLastUsed(TimeFilter):

    """Matches files by the time the file was last used.

    **`date_lastused` is only available on macOS!**

    Attributes:
        years (int): specify number of years
        months (int): specify number of months
        weeks (float): specify number of weeks
        days (float): specify number of days
        hours (float): specify number of hours
        minutes (float): specify number of minutes
        seconds (float): specify number of seconds
        mode (str):
            either 'older' or 'newer'. 'older' matches files / folders last used before
            the given time, 'newer' matches files / folders last used within the given
            time. (default = 'older')

    Returns:
        {date_lastused}: The datetime the files / folders were added.
    """

    filter_config: ClassVar[FilterConfig] = FilterConfig(
        name="date_lastused",
        files=True,
        dirs=True,
    )

    def __post_init__(self):
        if sys.platform != "darwin":
            raise EnvironmentError("date_added is only available on macOS")
        return super().__post_init__()

    def get_datetime(self, path: Path) -> datetime:
        return read_date_lastused(path)



from typing import ClassVar

from pydantic.config import ConfigDict
from pydantic.dataclasses import dataclass

from organize.action import ActionConfig
from organize.output import Output
from organize.resource import Resource
from organize.template import Template, render


@dataclass(config=ConfigDict(extra="forbid"))
class Echo:
    """Prints the given message.

    This can be useful to test your rules, especially in combination with placeholder
    variables.

    Attributes:
        msg (str): The message to print. Accepts placeholder variables.
    """

    msg: str = ""

    action_config: ClassVar[ActionConfig] = ActionConfig(
        name="echo",
        standalone=True,
        files=True,
        dirs=True,
    )

    def __post_init__(self):
        self._msg_templ = Template.from_string(self.msg)

    def pipeline(self, res: Resource, output: Output, simulate: bool):
        full_msg = render(self._msg_templ, res.dict())
        output.msg(res, full_msg, sender=self)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, NamedTuple, Protocol, runtime_checkable

if TYPE_CHECKING:
    from .output import Output
    from .resource import Resource


class FilterConfig(NamedTuple):
    name: str
    files: bool
    dirs: bool


@runtime_checkable
class HasFilterConfig(Protocol):
    filter_config: FilterConfig


class HasFilterPipeline(Protocol):
    def pipeline(self, res: Resource, output: Output) -> bool:
        ...  # pragma: no cover


@runtime_checkable
class Filter(HasFilterPipeline, HasFilterConfig, Protocol):
    def __init__(self, *args, **kwargs) -> None:
        # allow any amount of args / kwargs for BaseModel and dataclasses.
        ...  # pragma: no cover


class Not:
    def __init__(self, filter: Filter):
        self.filter = filter
        self.filter_config = self.filter.filter_config

    def pipeline(self, res: Resource, output: Output) -> bool:
        return not self.filter.pipeline(res=res, output=output)

    def __repr__(self):
        return f"Not({self.filter})"


class All:
    def __init__(self, *filters: Filter):
        self.filters = filters

    def pipeline(self, res: Resource, output: Output) -> bool:
        for filter in self.filters:
            try:
                match = filter.pipeline(res, output=output)
                if not match:
                    return False
            except Exception as e:
                output.msg(res=res, level="error", msg=str(e), sender=filter)
                logging.exception(e)
                return False
        return True


class Any:
    def __init__(self, *filters: Filter):
        self.filters = filters

    def pipeline(self, res: Resource, output: Output) -> bool:
        result = False
        for filter in self.filters:
            try:
                match = filter.pipeline(res, output=output)
                if match:
                    result = True
            except Exception as e:
                output.msg(res=res, level="error", msg=str(e), sender=filter)
                logging.exception(e)
        return result



import shutil
from typing import ClassVar, Literal

from pydantic.config import ConfigDict
from pydantic.dataclasses import dataclass

from organize.action import ActionConfig
from organize.output import Output
from organize.resource import Resource
from organize.template import Template, render

from .common.conflict import ConflictMode, resolve_conflict
from .common.target_path import prepare_target_path


@dataclass(config=ConfigDict(coerce_numbers_to_str=True, extra="forbid"))
class Copy:

    """Copy a file or dir to a new location.

    If the specified path does not exist it will be created.

    Attributes:
        dest (str):
            The destination where the file / dir should be copied to.
            If `dest` ends with a slash, it is assumed to be a target directory
            and the file / dir will be copied into `dest` and keep its name.

        on_conflict (str):
            What should happen in case **dest** already exists.
            One of `skip`, `overwrite`, `trash`, `rename_new` and `rename_existing`.
            Defaults to `rename_new`.

        rename_template (str):
            A template for renaming the file / dir in case of a conflict.
            Defaults to `{name} {counter}{extension}`.

        autodetect_folder (bool):
            In case you forget the ending slash "/" to indicate copying into a folder
            this settings will handle targets without a file extension as folders.
            If you really mean to copy to a file without file extension, set this to
            false.
            Defaults to True.

        continue_with (str) = "copy" | "original":
            Continue the next action either with the path to the copy or the path the
            original.
            Defaults to "copy".

    The next action will work with the created copy.
    """

    dest: str
    on_conflict: ConflictMode = "rename_new"
    rename_template: str = "{name} {counter}{extension}"
    autodetect_folder: bool = True
    continue_with: Literal["copy", "original"] = "copy"

    action_config: ClassVar[ActionConfig] = ActionConfig(
        name="copy",
        standalone=False,
        files=True,
        dirs=True,
    )

    def __post_init__(self):
        self._dest = Template.from_string(self.dest)
        self._rename_template = Template.from_string(self.rename_template)

    def pipeline(self, res: Resource, output: Output, simulate: bool):
        assert res.path is not None, "Does not support standalone mode"
        rendered = render(self._dest, res.dict())

        # fully resolve the destination for folder targets and prepare the folder
        # structure
        dst = prepare_target_path(
            src_name=res.path.name,
            dst=rendered,
            autodetect_folder=self.autodetect_folder,
            simulate=simulate,
        )

        # Resolve conflicts before copying the file to the destination
        skip_action, dst = resolve_conflict(
            dst=dst,
            res=res,
            conflict_mode=self.on_conflict,
            rename_template=self._rename_template,
            simulate=simulate,
            output=output,
        )

        if skip_action:
            return

        output.msg(res=res, msg=f"Copy to {dst}", sender=self)
        res.walker_skip_pathes.add(dst)
        if not simulate:
            if res.is_dir():
                shutil.copytree(src=res.path, dst=dst)
            else:
                shutil.copy2(src=res.path, dst=dst)

        # continue with either the original path or the path to the copy
        if self.continue_with == "copy":
            res.path = dst



import collections
import fnmatch
import json
import logging
import os
import subprocess
from datetime import date, datetime, timedelta
from functools import lru_cache
from pathlib import Path
from typing import Any, ClassVar, DefaultDict, Dict, Optional, Union

import exifread
from pydantic import BaseModel
from rich import print

from organize.filter import FilterConfig
from organize.output import Output
from organize.resource import Resource

ExifStrDict = Dict[str, Dict[str, str]]
ExifValue = Union[str, datetime, date, timedelta]
ExifDict = Dict[str, Dict[str, ExifValue]]
ExifDefaultDict = DefaultDict[str, DefaultDict[str, ExifValue]]

ORGANIZE_EXIFTOOL_PATH = os.environ.get("ORGANIZE_EXIFTOOL_PATH", "")


@lru_cache(maxsize=1)
def exiftool_available() -> bool:
    # don't use exiftool if user blanked the env path
    if not ORGANIZE_EXIFTOOL_PATH:
        return False

    # check whether the given path is executable
    try:
        subprocess.check_call(
            [ORGANIZE_EXIFTOOL_PATH, "-ver"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.STDOUT,
        )
        return True
    except subprocess.CalledProcessError:
        logging.warning("exiddftool not available. Falling back to exifread library.")
        return False


def group_keys_by_splitting(
    data: Dict[str, Any],
    delimiter: str = " ",
) -> Dict[str, Dict[str, Any]]:
    """
    >>> group_keys_by_splitting({"cat a": 1, "cat b": 2, "x c": 1, "x d": 2})
    {"cat": {"a": 1, "b": 2}, "x": {"c": 1, "d": 2}}
    """
    result: DefaultDict[str, Any] = collections.defaultdict(dict)
    for k, v in data.items():
        if delimiter in k:
            category, field = k.split(delimiter, maxsplit=1)
            result[category][field] = v
        else:
            result[k] = v
    return dict(result)


def lowercase_keys_recursive(data) -> Dict:
    if isinstance(data, Dict):
        return {k.lower(): lowercase_keys_recursive(v) for k, v in data.items()}
    return data


def parse_date_value(value: str) -> Union[datetime, date, str]:
    """
    Parse datetime or date values (e.g. image.datetime, exif.datetimeoriginal, ...)
    Exiftool often gives date and time values for keys with only "date" in their name
    so we have to check both in order to preserve this information.
    """
    try:
        return datetime.strptime(value[:19], "%Y:%m:%d %H:%M:%S")
    except ValueError:
        pass
    try:
        return datetime.strptime(value[:10], "%Y:%m:%d").date()
    except ValueError:
        pass
    return value


def parse_offset_value(value: str) -> Union[timedelta, str]:
    """
    Parse offset values (e.g. exif.offsettimeoriginal, exif.offsettimedigitized)
    Supports formats "+HHMM" or "+HH:MM[:SS]" or "UTC+HH:MM[:SS]"
    """
    try:
        if value[:3].upper() == "UTC":
            value = value[3:]  # remove UTC
        return datetime.strptime(value.replace(":", ""), "%z").utcoffset() or timedelta(
            seconds=0
        )
    except (ValueError, TypeError):
        return value


def convert_value(key: str, value: str) -> ExifValue:
    _key = key.lower()
    if "date" in _key:
        return parse_date_value(value)
    if "offset" in _key:
        return parse_offset_value(value)
    return value


def convert_recursive(data):
    result = dict()
    for k, v in data.items():
        if isinstance(v, Dict):
            result[k] = convert_recursive(v)
        else:
            result[k] = convert_value(k, v)
    return result


def exifread_read(path: Path) -> ExifStrDict:
    """
    Uses the `exifread` library to read the EXIF data
    """
    with path.open("rb") as f:
        data = exifread.process_file(fh=f, details=False, debug=False)
    # at this point data still contains exifread specific types like
    # Short / Ratio / ASCII which we now convert to a printable representation
    printable = {key: val.printable for (key, val) in data.items()}
    grouped = group_keys_by_splitting(printable)
    return grouped


def exiftool_read(path: Path) -> ExifStrDict:
    """
    Uses the `exiftool` tool by Phil Harvey to read the EXIF data
    """
    try:
        data_json = subprocess.check_output(
            (
                ORGANIZE_EXIFTOOL_PATH,
                "-j",
                "-g",
                "--fast",
                str(path),
            ),
            text=True,
        )
    except subprocess.CalledProcessError:
        return dict()

    # we pass a single filepath, so we are interested in the first element
    data: Dict = json.loads(data_json)[0]

    # if data contains no media specific EXIF keys we assume we don't have
    # any relevant EXIF data.
    if "Composite" not in data.keys():
        return dict()
    return data


def matches_tags(
    filter_tags: Dict[str, Optional[str]],
    data: Dict[str, Dict[str, str]],
) -> bool:
    if not data:
        return False
    for k, v in filter_tags.items():
        try:
            # step into the data dict by dotted notation
            data_value: Any = data
            for part in k.split("."):
                data_value = data_value[part]
            # if v is None it's enough for the key to exist in the data.
            # Otherwise we use a glob matcher to check for matches
            if v is not None and not fnmatch.fnmatch(data_value.lower(), v.lower()):
                return False
        except (KeyError, AttributeError):
            return False
    return True


class Exif(BaseModel):
    """Filter by image EXIF data

    The `exif` filter can be used as a filter as well as a way to get exif information
    into your actions.

    By default this library uses the `exifread` library. If your image format is not
    supported you can install `exiftool` (exiftool.org) and set the environment variable:

    ```
    ORGANIZE_EXIFTOOL_PATH="exiftool"
    ```

    organize will then use `exiftool` to extract the EXIF data.

    Exif fields which contain "datetime", "date" or "offsettime" in their fieldname
    will have their value converted to 'datetime.datetime', 'datetime.date' and
    'datetime.timedelta' respectivly.
    - `datetime.datetime` : exif.image.datetime, exif.exif.datetimeoriginal, ...
    - `datetime.date` : exif.gps.date, ...
    - `datetime.timedelta` : exif.exif.offsettimeoriginal, exif.exif.offsettimedigitized, ...

    Attributes:
        lowercase_keys (bool): Whether to lowercase all EXIF keys (Default: true)

    :returns:
        ``{exif}`` -- a dict of all the collected exif inforamtion available in the
        file. Typically it consists of the following tags (if present in the file):

        - ``{exif.image}`` -- information related to the main image
        - ``{exif.exif}`` -- Exif information
        - ``{exif.gps}`` -- GPS information
        - ``{exif.interoperability}`` -- Interoperability information
    """

    filter_tags: Dict
    lowercase_keys: bool = True

    filter_config: ClassVar[FilterConfig] = FilterConfig(
        name="exif",
        files=True,
        dirs=False,
    )

    def __init__(
        self,
        *args,
        filter_tags: Optional[Dict] = None,
        lowercase_keys: bool = True,
        **kwargs,
    ):
        # exif filter is used differently from other filters. The **kwargs are not
        # filter parameters but all belong into the filter_tags dictionary to filter
        # for specific exif tags.
        params = filter_tags or dict()
        params.update(kwargs)
        # *args are tags filtered without a value, like ["gps", "image.model"].
        for arg in args:
            params[arg] = None
        super().__init__(filter_tags=params, lowercase_keys=lowercase_keys)

    def pipeline(self, res: Resource, output: Output) -> bool:
        assert res.path is not None, "Does not support standalone mode"

        # gather the exif data in a dict
        if exiftool_available():
            data = exiftool_read(path=res.path)
        else:
            data = exifread_read(path=res.path)

        # lowercase keys if wanted
        if self.lowercase_keys:
            data = lowercase_keys_recursive(data)

        # convert strings to datetime objects where possible
        parsed = convert_recursive(data)

        res.vars[self.filter_config.name] = parsed
        return matches_tags(self.filter_tags, data)


if __name__ == "__main__":
    import sys

    # Usage:
    # python organize/filters/exif.py tests/resources/images-with-exif/3.jpg
    data = exifread_read(Path(sys.argv[1]))
    print("Exifread", data)
    if exiftool_available():
        data = exiftool_read(Path(sys.argv[1]))
        print("Exiftool", data)
    else:
        print("Exiftool not available")
